import re
import yaml
import traceback
import string
from typing import List, Optional
from pydantic import model_validator, ValidationInfo, field_validator, Field

from .base_model import BaseModel, IgnoredType
from .environment_variable import EnvironmentVariable
from .config import AppConfig
from .resource_use_existing import ResourceUseExistingSettings
from cnc.utils import clean_name_string

from cnc.logger import get_logger

log = get_logger(__name__)


class Environment(BaseModel):
    name: str
    # TODO: validate - no wildcard domains allowed here?
    custom_domains: Optional[List[str]] = []
    environment_variables: Optional[List[EnvironmentVariable]] = []
    config_file_path: Optional[str] = None
    config_data: Optional[dict] = {}
    image_tag: Optional[str] = "latest"
    # TODO: type should be set up for inclusion (static|preview|production)
    type: Optional[str] = "static"
    data: Optional[dict] = {}
    paused: Optional[bool] = False
    active_deployment: Optional[bool] = True
    database_password: Optional[EnvironmentVariable] = None
    provider: str
    raw_existing_resources: Optional[List[ResourceUseExistingSettings]] = Field(
        alias="existing_resources", default=[]
    )

    # ------------------------------
    # Parent relationships
    # (not meant to be set directly from configuration file)
    # ------------------------------
    config: Optional[IgnoredType] = IgnoredType()
    application: Optional[IgnoredType] = IgnoredType()
    collection: Optional[IgnoredType] = IgnoredType()

    # ------------------------------
    # Validators
    # ------------------------------
    @model_validator(mode="before")
    @classmethod
    def ensure_config_file_path_exists(cls, data: dict, info: ValidationInfo):
        """
        Sets default config_file_path if not set.
        See: Application.from_environments_yml for context
        """
        if not data.get("config_file_path"):
            data["config_file_path"] = info.context.get("config_file_path")
        return data

    @model_validator(mode="after")
    def ensure_config_data(self):
        """
        Loads config data from file if not already set
        """
        if not self.config_data:
            if self.config_file_path:
                with open(self.config_file_path) as coherence_yml_data:
                    self.config_data = yaml.safe_load(coherence_yml_data)
            else:
                log.debug("No config data and no config file path!")
                raise ValueError(f"No config data and no config file path for {self}!")

        self.config_data["environment"] = self

        try:
            self.config = AppConfig.model_validate(
                self.config_data, context={"provider": self.provider}
            )
        except Exception as e:
            log.warning(
                f"Cannot get AppConfig for environment! {traceback.format_exc()}"
            )
            raise e

        return self

    @model_validator(mode="after")
    def annotate_children(self):
        if self.config:
            for svc in self.config.services + self.config.settings.explicit_resources:
                svc.environment = self

        for variable in self.environment_variables:
            variable.environment = self

        if self.database_password:
            self.database_password.environment = self

        return self

    @field_validator("database_password", mode="before")
    def validate_database_password(cls, value: dict) -> dict:
        if value:
            value["name"] = "database_password"
        return value

    @field_validator("raw_existing_resources", mode="before")
    def validate_raw_existing_resources(cls, value: dict) -> list[dict]:
        return cls.convert_dict_keys_to_names(value)

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def is_static(self):
        return self.type in ["static", "production"]

    @property
    def is_production(self):
        return self.type == "production"

    @property
    def is_preview(self):
        return self.type == "preview"

    @property
    def services(self):
        if not hasattr(self, "_services"):
            self._services = []

        if not self._services and self.config:
            self._services += self.config.services
            self._services += self.config.settings.explicit_resources

        return self._services

    @property
    def web_services(self):
        return [s for s in self.services if s.is_web]

    @property
    def backend_services(self):
        if not self.services:
            return []

        return [s for s in self.services if s.is_backend]

    @property
    def frontend_services(self):
        if not self.services:
            return []

        return [s for s in self.services if s.is_frontend]

    @property
    def active(self):
        return not self.paused

    @property
    def database_resources(self):
        return [s for s in self.services if s.is_database]

    @property
    def cache_resources(self):
        return [s for s in self.services if s.is_cache]

    @property
    def message_queue_resources(self):
        return [s for s in self.services if s.is_message_queue]

    @property
    def object_storage_resources(self):
        return [s for s in self.services if s.is_object_storage]

    @property
    def filesystem_resources(self):
        return [s for s in self.services if s.is_filesystem]

    @property
    def load_database_snapshot(self):
        return any(
            [db for db in self.database_resources if db.settings.snapshot_file_path]
        )

    @property
    def service_domains(self):
        domains = []
        for service in self.services:
            if service.is_web:
                domains.append({"service_name": service.name, "domain": service.domain})
        return domains

    @property
    def domains(self):
        if self.collection.has_service_domains:
            return self.service_domains

        domains = [{"service_name": None, "domain": self.domain}]
        for domain in self.custom_domains:
            domains.append({"service_name": None, "domain": domain})

        return domains

    @property
    def domain(self):
        if not self.collection.has_service_domains:
            return f"{self.name}.{self.collection.base_domain}"

    @property
    def custom_ns_records(self):
        return self.data.get("custom_ns_records", {})

    @property
    def default_service(self):
        if self.frontend_services:
            return sorted(
                self.frontend_services,
                key=lambda x: len(x.settings.url_path),
            )[0]
        elif self.backend_services:
            return self.backend_services[0]

        return None

    @property
    def existing_resources(self):
        # order here is important so we return env-level first in the array!
        return self.raw_existing_resources + self.collection.existing_resources

    @property
    def instance_name(self):
        _name = (
            f"{self.collection.cloud_resource_namespace[:10]}-{self.application.name[-8:]}"
            f"-{self.collection.name[-9:]}-{self.name[-9:]}"
        )
        clean = clean_name_string(_name, truncate_len=39)

        if clean[0].lower() not in string.ascii_lowercase:
            clean = "c" + clean[1:]

        return re.sub("[\\W_]+$", "", clean)

    @property
    def managed_environment_items(self):
        return (
            self.config.managed_environment_variables
            + self.config.managed_environment_secrets
        )

    @property
    def environment_items(self):
        return self.environment_variables + self.managed_environment_items

    # ------------------------------
    # Instance methods
    # ------------------------------
    def __repr__(self):
        return (
            f"<Environment (name: {self.name} | collection: "
            f"{self.collection.name} | {[s.name for s in self.services]})>"
        )

    __str__ = __repr__

    def custom_ns_records_for(self, domain):
        return self.custom_ns_records.get(domain)

    def variable_by_name(self, name):
        for variable in self.environment_variables:
            if variable.name == name:
                return variable
        log.debug(f"{name} var not found for {self}")

    def service_by_name(self, name):
        for service in self.services:
            if service.name == name:
                return service
        log.debug(f"{name} service not found for {self}")



import typer
from cnc.models import ProvisionStageManager, DeployStageManager, BuildStageManager
from .telemetry import send_event

from cnc.logger import get_logger

log = get_logger(__name__)

app = typer.Typer()


@app.callback(invoke_without_command=True)
@app.command()
def render(
    ctx: typer.Context,
    stage: str = "provision",
    extra_context: dict = None,
):
    send_event("render.render")

    manager_map = {
        "build": BuildStageManager,
        "provision": ProvisionStageManager,
        "deploy": DeployStageManager,
    }

    if not manager_map.get("stage"):
        log.error(f"No manager for {stage}")
        raise typer.Exit(code=1)

    # get manager
    # render template with added context

    raise typer.Exit()



import os

import rudderstack.analytics as rudder_analytics
import machineid

rudder_analytics.write_key = "2eEyoetBjPLeYUOg8jRfOicBRiS"
rudder_analytics.dataPlaneUrl = "https://withcoherepvm.dataplane.rudderstack.com"
rudder_analytics.debug = True
rudder_analytics.gzip = True

# default to enabling telemetry for now
TELEMETRY_ENABLED = (
    False
    if os.environ.get("CNC_TELEMETRY_DISABLED")
    in [1, "1", "true", "True", "TRUE", "yes", "Y", "y", "Yes", "YES"]
    else True
)
CURRENT_COMMAND = ""


def send_event(command_name: str):
    command_data = {
        "name": command_name,
    }
    if TELEMETRY_ENABLED:
        try:
            _id = machineid.id()
        except Exception as e:
            print(f"Cannot get machine ID: {e}")
            _id = None

        try:
            user_id = os.environ.get("CNC_USER_ID", _id or "UNKNOWN")
            print(f"Sending {command_data} to RS for {user_id}")
            rudder_analytics.track(user_id, "command", command_data)
        except Exception as e:
            print(f"Cannot send telemetry event: {e}")



import os
import subprocess

from .cycle_stage_base import EnvironmentTemplatedBase
from .stage import Stage

from cnc.logger import get_logger

log = get_logger(__name__)


class DeployStageManager(EnvironmentTemplatedBase):
    template_type = "deploy"

    def __init__(self, *args, webhook_url=None, webhook_token=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.webhook_url = webhook_url
        self.webhook_token = webhook_token

    def template_context(self, service):
        # cluster name, subnet, security group, load_database_snapshot (flag from CLI?)

        return {
            "deployer": self,
            "service": service,
            "environment": self.environment,
            "stage": Stage.model_validate({"name": "deploy"}),
            "load_database_snapshot": False,
            "environment_items": self.environment_items,
        }

    def render_scripts(self, service_names=None):
        context = self.template_context(None)
        if os.path.isfile(f"{self.config_files_path}/pre_deploy_functions.sh.j2"):
            self.write_template(
                "pre_deploy_functions.sh.j2",
                output_name=f"pre-deploy-{self.environment.name}-functions.sh",
                context=context,
            )
            self.write_template(
                "pre_deploy.sh.j2",
                output_name=f"pre-deploy-{self.environment.name}.sh",
                context=context,
            )
            self.scripts_to_run.append(f"pre-deploy-{self.environment.name}.sh")

        for service in self.environment.services:
            if service_names:
                if service.name not in service_names:
                    continue

            if service.settings.is_resource:
                continue

            context = self.template_context(service)
            context["render_template"] = self.write_template_with_context(service)

            self.write_template(
                "deploy_functions.sh.j2",
                output_name=f"deploy-{service.name}-functions.sh",
                context=context,
            )
            self.write_template(
                "main.sh.j2", output_name=f"deploy-{service.name}.sh", context=context
            )
            self.scripts_to_run.append(f"deploy-{service.name}.sh")

    def perform(self, should_cleanup=True, should_regenerate_config=True, debug=False):
        log.debug(f"Performing deploy for {self} @ {self.config_files_path}")
        if should_cleanup:
            self.cleanup()
        self.setup()
        if should_regenerate_config:
            self.render_scripts(service_names=self.service_tags.keys())
            log.debug(
                f"Done rendering deploy for {self} (svcs: {self.service_tags.keys()}), going to execute..."
            )

        if debug:
            for script in self.scripts_to_run:
                self.debug_template_output(script)

            log.debug(f"All done with debug for {self}")
            return

        for script in self.scripts_to_run:
            # self.debug_template_output("run-app.yml")

            try:
                # log.debug(f"Going to run {script}...")
                ret = subprocess.run(
                    ["bash", "-c", f"source {self.rendered_files_path}/{script}"],
                    executable="/bin/bash",
                )
                if ret.returncode != 0:
                    return ret.returncode

                # log.debug(f"Output from {script}: \n{ret.stdout}\n{ret.stderr}")
                log.info(f"All done with perform for {self}")
            except subprocess.CalledProcessError as e:
                # Handle any errors that occurred during script execution
                print(f"Error running the script: {e}")
                print("Error output:")
                print(e.stderr)
                return 1

        return 0



from unittest.mock import patch

from .base_test_class import CNCBaseTestCase
from cnc.models import (
    ToolboxManager,
    Application,
)
from cnc.logger import get_logger

log = get_logger(__name__)


class BaseToolboxTest(CNCBaseTestCase):
    fixture_name = "backend-1-service-1-db"
    environment_tag = "latest"
    env_data_filepath = "environments.yml"

    def setUp(self):
        super().setUp()
        self.app = Application.from_environments_yml(
            self.env_data_filepath,
        )
        self.collection = self.app.collections[0]
        self.collection._infra_outputs_cache = {}

        self.environment = self.collection.environments[0]
        self.service = self.environment.backend_services[0]

        self.toolbox = ToolboxManager(
            service=self.service,
            service_tags={self.service.name: self.environment_tag},
        )

        self.toolbox.setup()
        with patch.object(
            self.collection.__class__,
            "get_secret_value",
            return_value="secretvalue",
        ):
            self.toolbox.render_toolbox()

    def tearDown(self):
        super().tearDown()
        if hasattr(self, "toolbox"):
            self.toolbox.cleanup()

    def parse(self, script_name="main.sh"):
        with open(
            f"{self.toolbox.rendered_files_path}/{script_name}",
            "r",
        ) as file:
            return file.read()


class AWSToolboxSmokeTest(BaseToolboxTest):
    env_data_filepath = "environments_aws_ecs.yml"

    def test_toolbox_render(self):
        toolbox_script = self.parse()
        self.assertIn("aws ssm start-session", toolbox_script)

        dockerfile = self.parse("Dockerfile")
        self.assertIn(
            f"{self.service.image_for_tag(self.environment_tag)}",
            dockerfile,
        )


class GCPToolboxSmokeTest(BaseToolboxTest):
    env_data_filepath = "environments.yml"

    def test_toolbox_render(self):
        toolbox_script = self.parse()
        self.assertIn("compute ssh toolbox@", toolbox_script)

        dockerfile = self.parse("Dockerfile")
        self.assertIn(
            f"{self.service.image_for_tag(self.environment_tag)}",
            dockerfile,
        )



import re
import hashlib
import json
import string
from typing import List, Optional

from pydantic import model_validator, field_validator, Field

from .environment import Environment
from .base_model import BaseModel, IgnoredType
from .environment_variable import EnvironmentVariable
from .providers.amazon import AmazonAppServiceAccount
from .provisioner import ProvisionStageManager
from .resource_use_existing import ResourceUseExistingSettings
from cnc.utils import clean_name_string

from cnc.logger import get_logger

log = get_logger(__name__)


class IntegrationSettings(BaseModel):
    data: Optional[dict] = None


class EnvironmentCollection(BaseModel):
    name: str
    provided_base_domain: Optional[str] = Field(alias="base_domain", default=None)
    default: bool = False
    account_id: str
    environments: List[Environment]
    data: Optional[dict] = {}
    cloud_resource_namespace_override: Optional[str] = Field(
        alias="cloud_resource_namespace",
        default=None,
    )
    integrations: Optional[IntegrationSettings] = None
    database_password: Optional[EnvironmentVariable] = None
    existing_resources: Optional[List[ResourceUseExistingSettings]] = []
    collection_region: Optional[str] = Field(alias="region", default=None)

    # ------------------------------
    # Parent relationships
    # (not meant to be set directly from configuration file)
    # ------------------------------
    application: Optional[IgnoredType] = IgnoredType()

    # ------------------------------
    # Validators
    # ------------------------------
    @model_validator(mode="after")
    def annotate_children(self):
        for environment in self.environments:
            environment.collection = self

            for variable in environment.environment_variables:
                variable.collection = self

            if environment.database_password:
                environment.database_password.collection = self

        if self.database_password:
            self.database_password.collection = self

        return self

    @field_validator("existing_resources", mode="before")
    def validate_existing_resources(cls, value: dict) -> list[dict]:
        return cls.convert_dict_keys_to_names(value)

    @field_validator("cloud_resource_namespace_override", mode="before")
    def validate_cloud_resource_namespace(cls, value: str) -> str:
        assert value.startswith("c"), "cloud_resource_namespace must start with a c"

        return value[:11].replace("-", "")

    @field_validator("database_password", mode="before")
    def validate_database_password(cls, value: dict) -> dict:
        value["name"] = "database_password"
        return value

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def cloud_resource_namespace(self):
        return self.__class__.calculate_cloud_resource_namespace(
            self.account_id, self.application.name, self.name
        )

    @classmethod
    def calculate_cloud_resource_namespace(
        cls, account_id, application_name, collection_name
    ):
        _hash = hashlib.sha256()
        _hash.update(account_id.encode())
        _hash.update(application_name.encode())
        _hash.update(collection_name.encode())
        _hash = _hash.hexdigest()
        return f"c{_hash[:11]}"

    @property
    def instance_name(self):
        _name = (
            f"{self.cloud_resource_namespace[:12]}-{self.application.name[-12:]}"
            f"-{self.name[-13:]}"
        )
        clean = clean_name_string(_name, truncate_len=39)

        if clean[0].lower() not in string.ascii_lowercase:
            clean = "c" + clean[1:]

        return re.sub("[\\W_]+$", "", clean)

    @property
    def service_account(self):
        if self.application.provider_is_gcp:
            pass
            # return GoogleAppServiceAccount(collection=self)
        elif self.application.provider_is_aws:
            return AmazonAppServiceAccount(collection=self)

    @property
    def unique_id(self):
        _state = f"{self.cloud_resource_namespace}:{self.name}"

        for environment in self.environments:
            _state = _state + environment.name

            _data = environment.config_data.copy()
            del _data["environment"]

            _state = _state + json.dumps(_data, sort_keys=True)

        _hash = hashlib.sha256()
        _hash.update(_state.encode())
        _unique_id = _hash.hexdigest()

        return _unique_id

    @property
    def config_files_path(self):
        return f"{self.application.config_files_path}/{self.unique_id}"

    @property
    def valid_environments(self):
        "Includes all valid environments (active + paused)"
        return [e for e in self.environments if e.config]

    @property
    def active_environments(self):
        "Valid active environments"
        return [e for e in self.valid_environments if e.active]

    @property
    def has_active_deployments(self):
        return any([e for e in self.active_environments if e.active_deployment])

    @property
    def paused_environments(self):
        "Valid paused environments"
        return [e for e in self.valid_environments if e.paused]

    @property
    def environments_with_databases(self):
        environments = []
        for environment in self.valid_environments:
            if environment.database_resources:
                environments.append(environment)
        return environments

    @property
    def backend_services(self):
        "non-unique, used in infra to provision several envs"
        return self.all_services_for_type("backend")

    @property
    def frontend_services(self):
        "non-unique, used in infra to provision several envs"
        return self.all_services_for_type("frontend")

    @property
    def all_services(self):
        return self.all_services_for_type()

    @property
    def all_web_services(self):
        return self.all_services_for_type(["frontend", "backend"])

    @property
    def database_resources(self):
        return self.all_services_for_type("database")

    @property
    def cache_resources(self):
        return self.all_services_for_type("cache")

    @property
    def object_storage_resources(self):
        return self.all_services_for_type("object_storage")

    @property
    def message_queue_resources(self):
        return self.all_services_for_type("message_queue")

    @property
    def has_object_storage(self):
        return bool(self.object_storage_resources)

    @property
    def has_message_queues(self):
        return bool(self.message_queue_resources)

    @property
    def load_database_snapshot(self):
        return any([e for e in self.environments if e.load_database_snapshot])

    @property
    def default_service(self):
        for service in self.frontend_services + self.backend_services:
            if service.environment.active_deployment:
                return service

    @property
    def has_service_domains(self):
        return False

    @property
    def base_domain(self):
        if self.provided_base_domain:
            return self.provided_base_domain

    @property
    def domains(self):
        domains = []
        for environment in self.environments:
            domains.extend(environment.domains)
        return sorted(domains, key=lambda x: x.get("domain"))

    @property
    def domain_buckets(self):
        buckets = {}

        for domain_info in self.domains:
            domain = domain_info.get("domain", "")
            # 6 buckets = approx. 70 domains on gcp
            bucket = int((hashlib.sha256(domain.encode()).hexdigest()), 16) % 6

            if bucket not in buckets:
                buckets[bucket] = []

            buckets[bucket].append(domain)

        return buckets

    @property
    def needs_k8s(self):
        if self.application.flavor == "gke":
            return True

        for service in self.all_services_for_type():
            if service.is_backend and (
                service.settings.workers or service.settings.scheduled_tasks
            ):
                return True

    @property
    def provision_stage_manager(self):
        if not hasattr(self, "_provision_stage_manager"):
            self._provision_stage_manager = ProvisionStageManager(self)
        return self._provision_stage_manager

    @property
    def region(self):
        return self.collection_region or self.application.region

    @property
    def includes_production(self):
        return any([e for e in self.environments if e.is_production])

    # ------------------------------
    # Instance methods
    # ------------------------------
    def __repr__(self):
        return (
            f"<EnvironmentCollection ({self.name} | "
            f"{self.account_id}) [{len(self.environments)} envs]>"
        )

    __str__ = __repr__

    def all_services_for_type(self, service_type=None):
        "non-unique, returns all active services by service_type"
        _all = []
        for environment in self.active_environments:
            for service in environment.services:
                if service_type:
                    if isinstance(service_type, list):
                        if service.settings.type not in service_type:
                            continue
                    elif service.settings.type != service_type:
                        continue

                _all.append(service)

        return sorted(_all, key=lambda x: x.settings.unique_id)

    def infra_outputs(self, force_cache_refresh=False):
        _config = None

        if "infrastructure_outputs" not in self.data:
            try:
                if (not hasattr(self, "_infra_outputs_cache")) or force_cache_refresh:
                    log.debug(f"Going to get outputs for {self}: {self.data}")
                    _config = ProvisionStageManager(
                        self,
                        output_only=True,
                    )
                    if not _config.make_ready_for_use():
                        raise Exception("Make ready for use failed")

                    self._infra_outputs_cache = _config.output()
            except Exception as e:
                log.debug(f"Cannot get TF outputs for {self}: {e}")
                if not hasattr(self, "_infra_outputs_cache"):
                    self._infra_outputs_cache = {}
            finally:
                if _config:
                    _config.cleanup()
        else:
            self._infra_outputs_cache = {}

        infra_outputs = {}
        infra_outputs.update(self.data.get("infrastructure_outputs", {}))
        infra_outputs.update(self._infra_outputs_cache)

        return infra_outputs

    def get_terraform_output(self, output_name, force_cache_refresh=False):
        return (
            self.infra_outputs(force_cache_refresh=force_cache_refresh)
            .get(output_name, {})
            .get("value", "")
        )

    def generate_tf_assets(self, *args, **kwargs):
        return True

    def environment_by_name(self, environment_name):
        for environment in self.environments:
            if environment.name == environment_name:
                return environment



from .base_model import BaseModel, IgnoredType
from .environment_variable import EnvironmentVariable
from .config import AppConfig, InternalSettings
from .environment import Environment
from .environment_collection import EnvironmentCollection
from .application import Application
from .builder import BuildStageManager
from .deployer import DeployStageManager
from .provisioner import ProvisionStageManager
from .toolbox import ToolboxManager
from .resource_use_existing import ResourceUseExistingSettings



import re
import string
from typing import (
    List,
    Optional,
    Dict,
    Any,
    Union,
    Annotated,
)
from pydantic import (
    field_validator,
    Field,
    model_validator,
    ValidationInfo,
)
from jinja2 import Template

from cnc.constants import EnvironmentVariableTypes
from cnc.utils import clean_name_string
from ..base_model import BaseModel, IgnoredType
from .resource import BucketResourceSettings
from .settings import (
    FrontendServiceSettings,
    BackendServiceSettings,
    ProviderDeployResourceLimits,
)
from cnc.models.providers.amazon import (
    AWSDatabaseResourceSettings,
    AWSCacheResourceSettings,
)
from cnc.models.providers.google import (
    GCPDatabaseResourceSettings,
    GCPCacheResourceSettings,
)
from cnc.models.toolbox import ToolboxManager

from cnc.logger import get_logger

log = get_logger(__name__)


class BuildSettings(BaseModel):
    raw_dockerfile: Optional[str] = Field(alias="dockerfile", default=None)
    raw_context: Optional[str] = Field(alias="context", default=None)

    @property
    def context(self):
        return self.raw_context or "."

    @property
    def dockerfile(self):
        return f"{self.context}/{self.raw_dockerfile or 'Dockerfile'}"

    @property
    def dockerfile_is_default(self):
        return not self.raw_dockerfile


class DeployResources(BaseModel):
    limits: Optional[ProviderDeployResourceLimits] = Field(
        default_factory=ProviderDeployResourceLimits,
    )

    # ------------------------------
    # Validators
    # ------------------------------

    @model_validator(mode="before")
    def ensure_limits_provider(cls, data, info):
        if not data.get("limits"):
            data["limits"] = {}

        data["limits"]["provider"] = info.context.get("provider")

        return data


class DeploySettings(BaseModel):
    resources: Optional[DeployResources] = Field(default_factory=DeployResources)
    replicas: Optional[int] = 1

    @model_validator(mode="before")
    def ensure_deploy_resources_context(cls, data):
        # This is meant to ensure that validation
        # context is passed to child models
        if not data.get("resources"):
            data["resources"] = {}

        return data


ProviderDatabaseResourceSettings = Annotated[
    Union[GCPDatabaseResourceSettings, AWSDatabaseResourceSettings],
    Field(discriminator="provider"),
]

ProviderCacheResourceSettings = Annotated[
    Union[GCPCacheResourceSettings, AWSCacheResourceSettings],
    Field(discriminator="provider"),
]


class Service(BaseModel):
    name: str = Field(pattern=r"^\S+$")
    settings: Union[
        ProviderDatabaseResourceSettings,
        ProviderCacheResourceSettings,
        BucketResourceSettings,
        FrontendServiceSettings,
        BackendServiceSettings,
    ] = Field(alias="x-cnc", discriminator="type")
    build: Optional[BuildSettings] = None
    deploy: DeploySettings
    image: Optional[str] = None
    compose_environment: Dict[str, str] = Field(
        default_factory=dict, alias="environment"
    )
    command: List[str] = Field(default_factory=list)
    ports: Optional[List[str]] = []

    # ------------------------------
    # Parent relationships
    # ------------------------------
    config: Optional[IgnoredType] = IgnoredType()
    environment: Optional[IgnoredType] = Field(
        default_factory=IgnoredType,
        alias="env_parent_rel",
    )

    # ------------------------------
    # Validators
    # ------------------------------
    @model_validator(mode="before")
    def ensure_settings_provider(cls, data, info):
        settings_key = "x-cnc"
        if data.get("x-cnc"):
            settings_key = "x-cnc"
        elif data.get("settings"):
            settings_key = "settings"
        else:
            data[settings_key] = {}

        data[settings_key]["provider"] = info.context.get("provider")

        return data

    @model_validator(mode="before")
    def ensure_deploy_settings_context(cls, data):
        # This is meant to ensure that validation
        # context is passed to child models
        if not data.get("deploy"):
            data["deploy"] = {}

        return data

    @model_validator(mode="after")
    def annotate_children(self):
        self.settings.service = self

        return self

    @field_validator("command", mode="before")
    def validate_command(cls, value: Any) -> list[str]:
        if isinstance(value, str):
            return value.split()
        return value

    @model_validator(mode="before")
    @classmethod
    def default_build_if_no_image(cls, data: dict, info: ValidationInfo):
        if not (data.get("image") or data.get("build")):
            data["build"] = {}

        return data

    # @field_validator("compose_environment", mode="before")
    # def validate_compose_env(cls, value):
    #     if isinstance(value, list):
    #         value = key_equal_value_to_dict(value)
    # pydantic v2 will not "convert" values to strings
    # also need to do something like this for missing values:
    # - value missing
    # - try to get it from os.environ
    # - if not available ignore item entirely
    # return {k: str(v) for k, v in value.items()}

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def port(self):
        if self.ports:
            return self.ports[0].split(":")[-1]

        return "8080"

    @property
    def instance_name(self):
        if (
            hasattr(self.settings, "existing_instance_name")
            and self.settings.existing_instance_name
        ):
            return self.settings.existing_instance_name

        _name = (
            f"{self.settings.unique_id[:10]}-{self.application.name[:5]}-"
            f"{self.environment.collection.name[-5:]}-"
            f"{self.environment.name[-7:]}-{self.name[-7:]}"
        )

        clean = clean_name_string(_name, truncate_len=39)

        # instance names in gcp often have to start with a letter...
        if clean[0].lower() not in string.ascii_lowercase:
            clean = "c" + clean[1:]

        return re.sub("[\\W_]+$", "", clean)

    @property
    def workload_identity_pool(self):
        return f"serviceAccount:{self.environment.collection.account_id}.svc.id.goog[{self.instance_name}/{self.instance_name}]"

    @property
    def is_internal(self):
        return self.settings.internal is True

    @property
    def is_backend(self):
        return self.settings.type == "backend"

    @property
    def is_frontend(self):
        return self.settings.type == "frontend"

    @property
    def is_database(self):
        return self.settings.type == "database"

    @property
    def is_cache(self):
        return self.settings.type == "cache"

    @property
    def is_message_queue(self):
        return self.settings.type == "message_queue"

    @property
    def is_object_storage(self):
        return self.settings.type == "object_storage"

    @property
    def is_filesystem(self):
        return self.settings.type == "filesystem"

    @property
    def is_web(self):
        return self.settings.type in ["frontend", "backend"]

    @property
    def included_build_globs(self):
        if self.build.context == ".":
            # match everything
            default_globs = ["**/*"]
            _glob_tmpl = "{_}"
        else:
            default_globs = [f"{self.build.context}/**/*"]
            _glob_tmpl = "{self.build.context}/{_}"

        return [
            _glob_tmpl.format(_=_, self=self)
            for _ in self.settings.build_config.included_paths
        ] or default_globs

    @property
    def excluded_build_globs(self):
        if self.settings.repo_path == ".":
            _glob_tmpl = "{_}"
        else:
            _glob_tmpl = "{self.settings.repo_path}/{_}"

        return [
            _glob_tmpl.format(_=_, self=self)
            for _ in self.settings.build_config.excluded_paths
        ]

    @property
    def environment_outputs(self):
        # aliases are included by the filtered_environment_items method
        return self.filtered_environment_items(
            variable_type=EnvironmentVariableTypes.VARIABLE_TYPE_OUTPUT
        )

    @property
    def environment_secrets(self):
        # aliases are included by the filtered_environment_items method
        return self.config.managed_environment_secrets_for_service(
            service=self
        ) + self.filtered_environment_items(
            variable_type=EnvironmentVariableTypes.VARIABLE_TYPE_SECRET
        )

    @property
    def environment_variables(self):
        # aliases are included by the filtered_environment_items method
        return self.config.managed_environment_variables_for_service(
            service=self
        ) + self.filtered_environment_items(
            variable_type=EnvironmentVariableTypes.VARIABLE_TYPE_STANDARD
        )

    @property
    def environment_items(self):
        # includes all 3 types
        return (
            self.environment_outputs
            + self.environment_secrets
            + self.environment_variables
        )

    @property
    def insecure_environment_items(self):
        # this does not include secrets, meant for deploy runtime config
        # e.g. ECS/cloud run, includes standard/outputs
        return self.environment_variables + self.environment_outputs

    @property
    def gcr_image_name(self):
        app_name = clean_name_string(self.config.environment.application.name, 12)
        service_name = clean_name_string(self.name, 15)
        return f"{app_name}-{service_name}"

    @property
    def domain(self):
        if self.is_web:
            # TODO: figure out how to offer subdomain routing here
            # TODO: do custom domains need to live here if so?
            if self.environment.collection.has_service_domains:
                return self.environment.collection.get_terraform_output(
                    f"{self.instance_name}_cloud_run_url"
                )

    @property
    def provider_links(self):
        metadata = self.environment.collection.application.flavor_metadata
        _templates = metadata.get("provider_links_templates", {})

        context = {"service": self}
        _links = [
            {
                "url": Template(_templates.get(self.settings.type, "")).render(context),
                "label": self.settings.type,
            }
        ]

        for worker in self.settings.workers:
            _context = context.copy()
            _context.update(
                {
                    "worker": worker,
                }
            )
            _links.append(
                {
                    "url": Template(_templates.get("worker", "")).render(_context),
                    "label": worker.name,
                    "type": "worker",
                }
            )

        for task in self.settings.scheduled_tasks:
            _context = context.copy()
            _context.update(
                {
                    "task": task,
                }
            )
            _links.append(
                {
                    "url": Template(_templates.get("task", "")).render(_context),
                    "label": task.name,
                    "type": "task",
                }
            )

        return _links

    @property
    def application(self):
        return self.environment.application

    @property
    def joined_command(self):
        if self.command:
            return " ".join(self.command)

    # ------------------------------
    # Instance methods
    # ------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name} (type: {self.settings.type})>"

    __str__ = __repr__

    def filtered_environment_items(self, **kwargs):
        kwargs["service_name"] = self.name
        return self.config.filtered_environment_items(**kwargs)

    def image_registry_url(self, run=False, region=None):
        if self.image:
            _split = self.image.split(":")
            if len(_split) == 2:
                image = _split[0]
            else:
                image = _split[0]
            return image

        _run_string = "-run" if run else ""
        region = region or self.config.environment.collection.region

        if self.environment.application.provider_is_gcp:
            return f"{self.config.environment.collection.region}-docker.pkg.dev/{self.environment.collection.account_id}/{self.config.environment.collection.instance_name}/{self.gcr_image_name}{_run_string}"
        elif self.environment.application.provider_is_aws:
            return f"{self.environment.collection.account_id}.dkr.ecr.{region}.amazonaws.com/{self.instance_name}"

    def image_tag(self, environment_tag):
        environment_tag = environment_tag or "latest"
        if self.settings.build_config.existing_image_tag_mode == "sha":
            return environment_tag

        image_tag = environment_tag
        if self.image:
            _split = self.image.split(":")
            if len(_split) == 2:
                image_tag = _split[1]

        return image_tag

    def image_for_tag(self, environment_tag="latest", region=None, run=False):
        if self.settings.is_resource:
            return self.image or ""
        return f"{self.image_registry_url(run=run, region=region)}:{self.image_tag(environment_tag=environment_tag)}"

    def log_stream_prefix(self, task_name="web"):
        return f"{self.instance_name}-{task_name}"

    def toolbox_manager(self, environment_tag="latest"):
        return ToolboxManager(self, service_tags={self.name: environment_tag})

    def cli_info(self):
        _data = {
            "name": self.name,
            "type": self.settings.type,
        }

        if self.is_frontend or self.is_backend:
            _data["url_path"] = self.settings.url_path

        return _data



from typing import ClassVar


class EnvironmentVariableTypes:
    VARIABLE_TYPE_SECRET: ClassVar[str] = "secret"
    VARIABLE_TYPE_OUTPUT: ClassVar[str] = "output"
    VARIABLE_TYPE_STANDARD: ClassVar[str] = "standard"
    VARIABLE_TYPE_ALIAS: ClassVar[str] = "alias"

    @classmethod
    def allowed_types(cls):
        return [
            cls.VARIABLE_TYPE_STANDARD,
            cls.VARIABLE_TYPE_SECRET,
            cls.VARIABLE_TYPE_OUTPUT,
            cls.VARIABLE_TYPE_ALIAS,
        ]


class EnvironmentVariableDestinations:
    VARIABLE_TYPE_NAME: ClassVar[str] = "name"
    VARIABLE_TYPE_FILE: ClassVar[str] = "file"

    @classmethod
    def allowed_types(cls):
        return [cls.VARIABLE_TYPE_NAME, cls.VARIABLE_TYPE_FILE]



from typing import Literal

from cnc.models.config.deploy_resource_limits import (
    DeployResourceLimits,
)
from cnc.logger import get_logger

log = get_logger(__name__)


class AWSDeployResourceLimits(DeployResourceLimits):
    provider: Literal["aws"]

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def cpu(self):
        return self.format_cpu(self.raw_cpu)

    @property
    def memory(self):
        return self.format_memory(self.raw_memory)

    # ------------------------------
    # Instance methods
    # ------------------------------

    def _format_memory_string(self, value):
        if value[-1:] == "M":
            try:
                return f"{(int(float(value[:-1])/256) + 1) * 256}"
            except Exception as e:
                log.warning(f"Unable to parse memory config for {self} | {e}")
                return "1GB"
        elif value[-1:] == "G":
            formatted_val = value.replace(" ", "")
            return f"{formatted_val}B"
        else:
            return "1GB"

    def format_memory(self, value):
        mem = self._format_memory_string(value)

        cpu_val = float(self._format_cpu_string(self.raw_cpu).split(" ")[0])
        # Get memory value and convert to MiB
        if mem[-1:] == "B":
            mem_val = float(mem[:-2]) * 1024
        else:
            mem_val = int(mem)

        # Ensure memory is enough to be compatible with cpu value
        # (see: https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task_definition_parameters.html#task_size)
        if cpu_val == 0.25 and mem_val < 512:
            return "512"
        elif cpu_val == 0.5 and mem_val < 1024:
            return "1024"
        elif cpu_val == 1 and mem_val < 2048:
            return "2048"
        elif cpu_val == 2 and mem_val < 4096:
            return "4096"
        elif cpu_val == 4 and mem_val < 8192:
            return "8192"
        else:
            return mem

    def _format_cpu_string(self, value):
        if not value:
            return "0.5 vCPU"

        # Only allow cpu units (vCPU) for aws
        try:
            val = float(value)
            if (val <= 1) and (val % 0.25 == 0):
                return f"{val} vCPU"
            elif val < 5:
                val = int(val) + (int(val) % 2)
                return f"{val} vCPU"
            elif val <= 16 and val % 2 == 0:
                return f"{val} vCPU"
            else:
                return "0.5 vCPU"
        except Exception:
            return "0.5 vCPU"

    def format_cpu(self, value):
        cpu = self._format_cpu_string(value)

        cpu_val = float(cpu.split(" ")[0])
        mem = self._format_memory_string(self.raw_memory)
        if mem[-1:] == "B":
            mem_val = float(mem[:-2]) * 1024
        else:
            mem_val = int(mem)

        # Ensure cpu is enough to be compatible with memory value
        # (see: https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task_definition_parameters.html#task_size)
        if mem_val > 16384 and cpu_val < 4:
            return "4 vCPU"
        elif mem_val > 8192 and cpu_val < 2:
            return "2 vCPU"
        elif mem_val > 4096 and cpu_val < 1:
            return "1 vCPU"
        elif mem_val > 2048 and cpu_val < 0.5:
            return "0.5 vCPU"
        else:
            return cpu
